/* Codeset bitstream encoding and decoding */

class Codeset {
	constructor(codetable) {
		/* build binary tree from input codetable */
		/* sort array */
		codetable.sort(function(a, b) {
			a = a[1];
			b = b[1];
			return b-a;
		});
		/* create binpath_table */
		this.binpath_table = new Object();
		for (var i in codetable) {
			this.binpath_table[codetable[i][0]] = '';
		}
		/* build binary tree */
		this.build_tree(codetable);

    /*
    console.log("New table " + Object.keys(this.binpath_table).length + " characters")
    for (i in this.binpath_table) {
      console.log(i + " - " + this.binpath_table[i]);
    }
    */
	}
	
	build_tree(codetable) {
		/* if codetable has one item, return */
		if (codetable.length == 1) {
			return
		}
		/* start with left half empty and all items in the right half */
		var left_half = [];
		var right_half = codetable.slice(0);
		var sum_left = 0, sum_right = 0;
		for (var i in codetable) {
			sum_right += codetable[i][1];
		}
		/* move items from right to left till left half has bigger sum of weights  than right half*/
		while (sum_left < sum_right) {
				left_half.push(right_half[0]);
				sum_left += right_half[0][1];
				sum_right -= right_half[0][1];
				right_half.shift();
		}
		/* append 0 to left half items, 1 to right half items and call build_tree on both halves */
		for (i in left_half) {
			this.binpath_table[left_half[i][0]] += "0";
		}
		for (i in right_half) {
			this.binpath_table[right_half[i][0]] += "1";
		}
		this.build_tree(left_half);
		this.build_tree(right_half);
	}
	
	encode(string) {
		/* return bitstream */
		var bitstream = "";
		for (var i in string) {
			if (string[i] in this.binpath_table) {
				bitstream += this.binpath_table[string[i]];
			}
		}
		return bitstream;
	}
	
	decode(bitstream) {
		/* return string */
		/* add bits to a buffer until match found, then reset buffer and try match again.*/
		var string = "", buffer = "", m = "";
		for (var i in bitstream) {
			buffer += bitstream[i];
			if (m = this.binpath_match(buffer)) {
				string += m;
				buffer = "";
			}
		}
		/* if no more bits and no match, add zeroes till match */
		while (buffer) {
			buffer += "0";
			if (m = this.binpath_match(buffer)) {
				string += m;
				buffer = "";
			}
		}
		
		return string;
	}
	
	binpath_match(pattern) {
		for (var key in this.binpath_table) {
			if (this.binpath_table[key] == pattern) {
				return key;
			}
		}
		return '';
	}
	
	charset_regex() {
		var regex = ''
		for (var key in this.binpath_table) {
			regex += key;
		}
    /* escape special characters */
    return regex.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g, "\\$&"); 
	}
}
